@@ -209,10 +209,12 @@ sendClientData cparams ctx = sendCertificate >> sendClientKeyXchg >> sendCertifi
             sendPacket ctx $ Handshake [ClientKeyXchg ckx]
           where getCKX_DHE = do
                     xver <- usingState_ ctx getVersion
-                    (ServerDHParams dhparams serverDHPub) <- fromJust <$> usingHState ctx (gets hstServerDHParams)
-                    (clientDHPriv, clientDHPub) <- generateDHE ctx dhparams
+                    serverParams <- fromJust <$> usingHState ctx (gets hstServerDHParams)
+                    (clientDHPriv, clientDHPub) <- generateDHE ctx (serverDHParamsToParams serverParams)
 
-                    let premaster = dhGetShared dhparams clientDHPriv serverDHPub
+                    let premaster = dhGetShared (serverDHParamsToParams serverParams)
+                                                clientDHPriv
+                                                (serverDHParamsToPublic serverParams)
                     usingHState ctx $ setMasterSecretFromPre xver ClientRole premaster
 
                     return $ CKX_DH clientDHPub
@@ -258,10 +260,8 @@ sendClientData cparams ctx = sendCertificate >> sendClientKeyXchg >> sendCertifi
                         _     -> return Nothing
 
                     -- Fetch all handshake messages up to now.
-                    msgs <- usingHState ctx $ B.concat <$> getHandshakeMessages
-                    (hashMethod, toSign) <- prepareCertificateVerifySignatureData ctx usedVersion malg msgs
-
-                    sigDig <- signatureCreate ctx malg hashMethod toSign
+                    msgs   <- usingHState ctx $ B.concat <$> getHandshakeMessages
+                    sigDig <- certificateVerifyCreate ctx usedVersion malg msgs
                     sendPacket ctx $ Handshake [CertVerify sigDig]
 
                 _ -> return ()
@@ -389,15 +389,13 @@ processServerKeyExchange ctx (ServerKeyXchg origSkx) = do
                 (c,_)           -> throwCore $ Error_Protocol ("unknown server key exchange received, expecting: " ++ show c, True, HandshakeFailure)
         doDHESignature dhparams signature signatureType = do
             -- TODO verify DHParams
-            expectedData <- generateSignedDHParams ctx dhparams
-            verified <- signatureVerify ctx signatureType expectedData signature
-            when (not verified) $ throwCore $ Error_Protocol ("bad " ++ show signatureType ++ " for dhparams", True, HandshakeFailure)
+            verified <- digitallySignDHParamsVerify ctx dhparams signatureType signature
+            when (not verified) $ throwCore $ Error_Protocol ("bad " ++ show signatureType ++ " for dhparams " ++ show dhparams, True, HandshakeFailure)
             usingHState ctx $ setServerDHParams dhparams
 
         doECDHESignature ecdhparams signature signatureType = do
             -- TODO verify DHParams
-            expectedData <- generateSignedECDHParams ctx ecdhparams
-            verified <- signatureVerify ctx signatureType expectedData signature
+            verified <- digitallySignECDHParamsVerify ctx ecdhparams signatureType signature
             when (not verified) $ throwCore $ Error_Protocol ("bad " ++ show signatureType ++ " for dhparams", True, HandshakeFailure)
             usingHState ctx $ setServerECDHParams ecdhparams
 
