@@ -41,6 +41,7 @@ module Network.TLS.Packet
     , decodePreMasterSecret
     , encodePreMasterSecret
     , encodeSignedDHParams
+    , encodeSignedECDHParams
 
     , decodeReallyServerKeyXchgAlgorithmData
 
@@ -51,6 +52,10 @@ module Network.TLS.Packet
     , generateServerFinished
 
     , generateCertificateVerify_SSL
+
+    -- * for extensions parsing
+    , getSignatureHashAlgorithm
+    , putSignatureHashAlgorithm
     ) where
 
 import Network.TLS.Struct
@@ -67,13 +72,11 @@ import Data.X509 (CertificateChainRaw(..), encodeCertificateChain, decodeCertifi
 import Network.TLS.Crypto
 import Network.TLS.MAC
 import Network.TLS.Cipher (CipherKeyExchangeType(..))
+import Network.TLS.Util.Serialization (os2ip,i2ospOf_)
 import Data.ByteString (ByteString)
 import qualified Data.ByteString as B
 import qualified Data.ByteString.Char8 as BC
 
-import qualified Crypto.Hash.SHA1 as SHA1
-import qualified Crypto.Hash.MD5 as MD5
-
 data CurrentParams = CurrentParams
     { cParamsVersion     :: Version                     -- ^ current protocol version
     , cParamsKeyXchgType :: Maybe CipherKeyExchangeType -- ^ current key exchange type
@@ -287,12 +290,25 @@ decodeClientKeyXchg cp = -- case  ClientKeyXchg <$> (remaining >>= getBytes)
         parseCKE CipherKeyExchange_DHE_RSA = parseClientDHPublic
         parseCKE CipherKeyExchange_DHE_DSS = parseClientDHPublic
         parseCKE CipherKeyExchange_DH_Anon = parseClientDHPublic
+        parseCKE CipherKeyExchange_ECDHE_RSA = parseClientECDHPublic
         parseCKE _                         = error "unsupported client key exchange type"
         parseClientDHPublic = CKX_DH . dhPublic <$> getInteger16
+        parseClientECDHPublic = do
+            len <- getWord8
+            _ <- getWord8 -- Magic number 4
+            let siz = fromIntegral len `div` 2
+            xb <- getBytes siz
+            yb <- getBytes siz
+            let x = os2ip xb
+                y = os2ip yb
+            return $ CKX_ECDH $ ecdhPublic x y siz
 
 decodeServerKeyXchg_DH :: Get ServerDHParams
 decodeServerKeyXchg_DH = getServerDHParams
 
+-- We don't support ECDH_Anon at this moment
+-- decodeServerKeyXchg_ECDH :: Get ServerECDHParams
+
 decodeServerKeyXchg_RSA :: Get ServerRSAParams
 decodeServerKeyXchg_RSA = ServerRSAParams <$> getInteger16 -- modulus
                                           <*> getInteger16 -- exponent
@@ -312,6 +328,10 @@ decodeServerKeyXchgAlgorithmData ver cke = toCKE
                 dhparams  <- getServerDHParams
                 signature <- getDigitallySigned ver
                 return $ SKX_DHE_DSS dhparams signature
+            CipherKeyExchange_ECDHE_RSA -> do
+                dhparams  <- getServerECDHParams
+                signature <- getDigitallySigned ver
+                return $ SKX_ECDHE_RSA dhparams signature
             _ -> do
                 bs <- remaining >>= getBytes
                 return $ SKX_Unknown bs
@@ -362,6 +382,11 @@ encodeHandshakeContent (ClientKeyXchg ckx) = do
     case ckx of
         CKX_RSA encryptedPreMaster -> putBytes encryptedPreMaster
         CKX_DH clientDHPublic      -> putInteger16 $ dhUnwrapPublic clientDHPublic
+        CKX_ECDH clientECDHPublic  -> do
+            let (x,y,siz) = ecdhUnwrapPublic clientECDHPublic
+            let xb = i2ospOf_ siz x
+                yb = i2ospOf_ siz y
+            putOpaque8 $ B.concat [B.singleton 4,xb,yb]
 
 encodeHandshakeContent (ServerKeyXchg skg) =
     case skg of
@@ -369,6 +394,7 @@ encodeHandshakeContent (ServerKeyXchg skg) =
         SKX_DH_Anon params     -> putServerDHParams params
         SKX_DHE_RSA params sig -> putServerDHParams params >> putDigitallySigned sig
         SKX_DHE_DSS params sig -> putServerDHParams params >> putDigitallySigned sig
+        SKX_ECDHE_RSA params sig -> putServerECDHParams params >> putDigitallySigned sig
         SKX_Unparsed bytes     -> putBytes bytes
         _                      -> error ("encodeHandshakeContent: cannot handle: " ++ show skg)
 
@@ -457,14 +483,31 @@ putSignatureHashAlgorithm (h,s) =
     putWord8 (valOfType h) >> putWord8 (valOfType s)
 
 getServerDHParams :: Get ServerDHParams
-getServerDHParams = ServerDHParams <$> getDHParams <*> getDHPublic
-  where getDHParams = dhParams <$> getInteger16 -- p
-                               <*> getInteger16 -- g
-        getDHPublic = dhPublic <$> getInteger16 -- y(server)
+getServerDHParams = ServerDHParams <$> getBigNum16 <*> getBigNum16 <*> getBigNum16
 
 putServerDHParams :: ServerDHParams -> Put
-putServerDHParams (ServerDHParams dhparams dhpub) =
-    mapM_ putInteger16 $ dhUnwrap dhparams dhpub
+putServerDHParams (ServerDHParams p g y) = mapM_ putBigNum16 [p,g,y]
+
+getServerECDHParams :: Get ServerECDHParams
+getServerECDHParams = do
+    _ <- getWord8 -- ECParameters ECCurveType: curve name type, should be 3
+    w16 <- getWord16   -- ECParameters NamedCurve
+    mxy <- getOpaque16 -- ECPoint
+    let xy = B.drop 1 mxy
+        siz = B.length xy `div` 2
+        (xb,yb) = B.splitAt siz xy
+        x = os2ip xb
+        y = os2ip yb
+    return $ ServerECDHParams (ecdhParams w16) (ecdhPublic x y siz)
+
+putServerECDHParams :: ServerECDHParams -> Put
+putServerECDHParams (ServerECDHParams ecdhparams ecdhpub) = do
+    let (w16,x,y,siz) = ecdhUnwrap ecdhparams ecdhpub
+    putWord8 3    -- ECParameters ECCurveType: curve name type
+    putWord16 w16 -- ECParameters NamedCurve
+    let xb = i2ospOf_ siz x
+        yb = i2ospOf_ siz y
+    putOpaque8 $ B.concat [B.singleton 4,xb,yb] -- ECPoint
 
 getDigitallySigned :: Version -> Get DigitallySigned
 getDigitallySigned ver
@@ -516,8 +559,8 @@ type PRF = Bytes -> Bytes -> Int -> Bytes
 generateMasterSecret_SSL :: Bytes -> ClientRandom -> ServerRandom -> Bytes
 generateMasterSecret_SSL premasterSecret (ClientRandom c) (ServerRandom s) =
     B.concat $ map (computeMD5) ["A","BB","CCC"]
-  where computeMD5  label = MD5.hash $ B.concat [ premasterSecret, computeSHA1 label ]
-        computeSHA1 label = SHA1.hash $ B.concat [ label, premasterSecret, c, s ]
+  where computeMD5  label = hash MD5 $ B.concat [ premasterSecret, computeSHA1 label ]
+        computeSHA1 label = hash SHA1 $ B.concat [ label, premasterSecret, c, s ]
 
 generateMasterSecret_TLS :: PRF -> Bytes -> ClientRandom -> ServerRandom -> Bytes
 generateMasterSecret_TLS prf premasterSecret (ClientRandom c) (ServerRandom s) =
@@ -539,8 +582,8 @@ generateKeyBlock_SSL :: ClientRandom -> ServerRandom -> Bytes -> Int -> Bytes
 generateKeyBlock_SSL (ClientRandom c) (ServerRandom s) mastersecret kbsize =
     B.concat $ map computeMD5 $ take ((kbsize `div` 16) + 1) labels
   where labels            = [ uncurry BC.replicate x | x <- zip [1..] ['A'..'Z'] ]
-        computeMD5  label = MD5.hash $ B.concat [ mastersecret, computeSHA1 label ]
-        computeSHA1 label = SHA1.hash $ B.concat [ label, mastersecret, s, c ]
+        computeMD5  label = hash MD5 $ B.concat [ mastersecret, computeSHA1 label ]
+        computeSHA1 label = hash SHA1 $ B.concat [ label, mastersecret, s, c ]
 
 generateKeyBlock :: Version -> ClientRandom -> ServerRandom -> Bytes -> Int -> Bytes
 generateKeyBlock SSL2  = generateKeyBlock_SSL
@@ -555,8 +598,8 @@ generateFinished_TLS prf label mastersecret hashctx = prf mastersecret seed 12
 
 generateFinished_SSL :: Bytes -> Bytes -> HashCtx -> Bytes
 generateFinished_SSL sender mastersecret hashctx = B.concat [md5hash, sha1hash]
-  where md5hash  = MD5.hash $ B.concat [ mastersecret, pad2, md5left ]
-        sha1hash = SHA1.hash $ B.concat [ mastersecret, B.take 40 pad2, sha1left ]
+  where md5hash  = hash MD5 $ B.concat [ mastersecret, pad2, md5left ]
+        sha1hash = hash SHA1 $ B.concat [ mastersecret, B.take 40 pad2, sha1left ]
 
         lefthash = hashFinal $ flip hashUpdateSSL (pad1, B.take 40 pad1)
                              $ foldl hashUpdate hashctx [sender,mastersecret]
@@ -579,6 +622,13 @@ generateServerFinished ver
 generateCertificateVerify_SSL :: Bytes -> HashCtx -> Bytes
 generateCertificateVerify_SSL = generateFinished_SSL ""
 
-encodeSignedDHParams :: ClientRandom -> ServerRandom -> ServerDHParams -> Bytes
-encodeSignedDHParams cran sran dhparams = runPut $
+encodeSignedDHParams :: ServerDHParams -> ClientRandom -> ServerRandom -> Bytes
+encodeSignedDHParams dhparams cran sran = runPut $
     putClientRandom32 cran >> putServerRandom32 sran >> putServerDHParams dhparams
+
+-- Combination of RFC 5246 and 4492 is ambiguous.
+-- Let's assume ecdhe_rsa and ecdhe_dss are identical to
+-- dhe_rsa and dhe_dss.
+encodeSignedECDHParams :: ServerECDHParams -> ClientRandom -> ServerRandom -> Bytes
+encodeSignedECDHParams dhparams cran sran = runPut $
+    putClientRandom32 cran >> putServerRandom32 sran >> putServerECDHParams dhparams
